/*
**	deconv.c
*/
#include "deconv.h"
#define T 1E-8
#define U (V/2)
//----------------static variables---------------------------------------------
int accl= 2;	// accellation value
int estm= 3;	// 0:alg0 (error deliver to estimated image by psf)
		// 1:alg1 (err ddeliver to est by psf*est with normalization)
		// 2:lucy // Lucy's algorithm 
		// 3:unified modification of est
int psfm= PSFM;	// 1:lucy rate(chi)^accl * est(xi) (Lucy's PSF modification)
		// 2:adl  not calc 	(deliver same error to est and psf)
		// 3:smpl error(chi)^accl (>th) * est(xi)
		// 4:smpl sqrt(sqrt())
int avge=1;	// 1: average est
int dispf= 1;	// display observed and psf window
int cutof= 0;	// cut_out utililty
int prm= 0;	// print PSFmatrix
int pre= 0;	// print error (rms)
int iloop= LOOP;
int hloop= HLOOP;
int cloop= 20;
int use_gamma= 1;
extern int grid;
int peak_pos_x;
int peak_pos_y;

#ifdef GET_PSF
int gpsf= 1;
#else
int gpsf= 0;
#endif
Dt	mat2[V][H];	// input h(l,k)
//----------------------initial PSF's------------------------------------------
int cx=  H/2; int cy= V/2;
int ix= -H/2; int ax= H-1-H/2;
int iy= -V/2; int ay= V-1-V/2;
#ifdef HIR
int ixa[5]={    -H/2,     -H2/2,     -H3/2,     -H4/2,     -H5/2};
int iya[5]={    -V/2,     -V2/2,     -V3/2,     -V4/2,     -V5/2};
int axa[5]={ H-1-H/2, H2-1-H2/2, H3-1-H3/2, H4-1-H4/2, H5-1-H5/2};
int aya[5]={ V-1-V/2, V2-1-V2/2, V3-1-V3/2, V4-1-V4/2, V5-1-V5/2};
#endif
#ifdef IIR2
Dt inp[SY][SX];
Dt flt[SY][SX];
Dt err[SY][SX];
Dt rat[SY][SX];
Dt est[SY][SX];
Dt tmp[SY][SX];
#ifdef BEST
Dt est0[SY][SX];
#endif
#endif
//------------------------ functions ------------------------------------------
Dt gmm= GMM;
Dt gamma_(Dt v){	// ~ pow(x/255,2.222)
	v= (v - OFFSET)/255.;
	if(v<0.) v= 0.; if(v>1.) v= 1.;

	if(use_gamma){
#ifdef GAMMA
	if(v< LID*GRD) v/= GRD;
	else v= pow((v+EXX)/(1+EXX),gmm);
#else
	v= pow(v,gmm);
#endif
	}
	if(v<0.) v= 0.; if(v>1.) v= 1.;
	return v;
}

Dt igamma(Dt v){	// ~ pow(x,0.45)*255
	if(v<0.) v= 0.; if(v>1.) v= 1.;

	if(use_gamma){
#ifdef GAMMA
	if(v< LID) v*= GRD;
	else v= (1+EXX)*pow(v,1./gmm)-EXX;
#else
	v= pow(v,1./gmm);
#endif
	}
	v= 255*v + OFFSET +.5;
	if(v<0.)v= 0.; if(v>255.)v= 255.;
	return v;
}

void
pr_mat(Dt m[V][H]) {
	int i,j;
	for(j=0;j<V;j++){
	for(i=0;i<H;i++) {
		if(m[j][i]==0) printf(" -");
		else printf("%2d",(int)(m[j][i]*V*H));
	}
	printf("\n");
	}
	printf("\n");
	//wtp(H,V,m);
}
#ifdef HIR
void
pr_mat_hir(int h, int v, Dt m[v][h])
{
	int i,j;
	for(j=0;j<v;j++){
	for(i=0;i<h;i++){
		if(m[j][i]==0) printf(" -");
		else printf("%2d",(int)(m[j][i]*v*h));
	}
	printf("\n");
	}
	printf("\n");
}
#endif

//----------------------- small array funcs ------------------------------------
void clr(Dt a[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]= 0; }
void no_minus(Dt a[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) if(a[j][i]<0) a[j][i]= 0; }
void plus(Dt a[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]= fabs(a[j][i]); }
void sqr(Dt a[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]= sqrt(a[j][i]); }
void mls(Dt a[V][H],Dt b) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]*= b; }
void add(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]+= b[j][i]; }
void add2(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]+= b[j][i]/3; }
void sub(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]-= b[j][i]; }
void mlt(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]*= b[j][i]; }
void dvd(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]/= b[j][i]; }
void cpy(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]=  b[j][i]; }
void subs(Dt a[V][H],int c) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) { a[j][i]-= c; if(a[j][i]<0)a[j][i]= 0;} }
void avg(Dt a[V][H],Dt b[V][H]) { int i,j; for(j=0;j<V;j++) for(i=0;i<H;i++) a[j][i]= (a[j][i]+b[j][i])/2; }
Dt  sum(Dt a[V][H]) { int i,j; Dt s= 0; for(j=0;j<V;j++) for(i=0;i<H;i++) s+= a[j][i]; return s; }
void clrl(Dt* a,      int l)	{ int i; for(i=0;i<l;i++) a[i]= 0; }
void addl(Dt* a,Dt* b,int l)	{ int i; for(i=0;i<l;i++) a[i]+= b[i]; }
void mltl(Dt* a,Dt* b,int l)	{ int i; for(i=0;i<l;i++) a[i]*= b[i]; }
void mlsl(Dt* a,Dt s, int l)	{ int i; for(i=0;i<l;i++) a[i]*= s; }
void intpl(Dt* a,Dt* b, Dt s, int l)	{ int i; for(i=0;i<l;i++) a[i]+= (b[i]-a[i])*s; }
void cpyl(Dt* a,Dt* b,int l)	{ int i; for(i=0;i<l;i++) a[i]= b[i]; }
void avgl(Dt* a,Dt* b,int l)	{ int i; for(i=0;i<l;i++) a[i]= (a[i]+b[i])/2; }

//------ normalize (only from ix to ax, from iy to ay)-----------------
int
norm1(Dt m[V][H], double c) // normalize sum of mat to 1.
{
	int i,j; double sum= 0;
	for(j=iy;j<=ay;j++)
	for(i=ix;i<=ax;i++) sum+= m[j+U][i+M];
	sum/= c;
	if(sum==0.) return 1;
	for(j=iy;j<=ay;j++)
	for(i=ix;i<=ax;i++) m[j+U][i+M]/= sum;
	return 0;
}
int
norm_hir(int h, int v, Dt m[v][h])
{
	int i,j; Dt sum= 0;
	//for(j=0;j<v;j++) for(i=0;i<h;i++) m[j][i]= MAX(0,m[j][i]);
	for(j=0;j<v;j++) for(i=0;i<h;i++) sum+= m[j][i];
	if(sum==0.) fprintf(stderr,"norm_hir: psf wrong\n");
	else for(j=0;j<v;j++) for(i=0;i<h;i++) m[j][i]/= sum;
	if(prm)pr_mat_hir(h,v,m);
	//wtp(h,v,m);
	return 0;
}

void
shrink_img(int ex,int ey, Dt* b0, Dt* a0) {
	int i,j;
	Dt (*b)[ex]= (Dt (*)[ex]) b0;
	Dt (*a)[ex*2]= (Dt (*)[ex*2]) a0;
#ifdef SHRINK_FIL
	int k,l;
#define Z 1E-6
	Dt t,c[ey*2][ex];
	for(j=0;j<ey*2;j++){
		c[j][0]= a[j][0];
		c[j][1]= (a[j][1]+2*a[j][2]+a[j][3])/4;
		for(i=2;i<ex-2;i++){ k= i*2;
			t=(8*a[j][k]+5*(a[j][k-1]+a[j][k+1])-(a[j][k-3]+a[j][k+3]))/16;
			c[j][i]=(t<Z)?Z:t;
		}
		c[j][i]= (a[j][i*2-1]+2*a[j][i*2]+a[j][i*2+1])/4;
		i++; c[j][i]= a[j][i*2];
	}
	for(i=0;i<ex;i++){
		b[0][i]= c[0][i];
		b[1][i]= (c[1][i]+2*c[2][i]+c[3][i])/4;
		for(j=2;j<ey-2;j++){ l= j*2;
			t=(8*c[l][i]+5*(c[l-1][i]+c[l+1][i])-(c[l-3][i]+c[l+3][i]))/16;
			b[j][i]=(t<Z)?Z:t;
		}
		b[j][i]= (c[2*j-1][i]+2*c[j*2][i]+c[j*2+1][i])/4;
		j++; b[j][i]= c[j*2][i];
	}
#else
	for(j=0;j<ey;j++) for(i=0;i<ex;i++) b[j][i]= SPEL(a);
#endif
}

void
expand_img(int ex, int ey, Dt* a0, Dt* b0) { int i,j;
	Dt (*a)[ex]=  (Dt(*)[ex]) a0;
	Dt (*b)[ex/2]= (Dt(*)[ex/2]) b0;
#ifdef EXP_IMG_FIL
	// original position preserving expander for shrinked by 4-pel average
	// copy to 4 pels and 121 filter(x and y direction)
	Dt (*c)[ex]= malloc(ex*ey*sizeof(Dt));
	Dt (*d)[ex]= malloc(ex*ey*sizeof(Dt));
	for(j=0;j<ey/2;j++) 
	for(i=0;i<ex/2;i++)
		c[2*j][2*i]= c[2*j][2*i+1]= c[2*j+1][2*i]= c[2*j+1][2*i+1]= b[j][i];

	for(j=0;j<ey;j++) {
		d[j][0]= c[j][0];
		for(i=1;i<ex-1;i++) d[j][i]= (c[j][i-1]+2*c[j][i]+c[j][i+1])/4;
		d[j][i]= c[j][i];
	}
	for(i=0;i<ex;i++) {
		a[0][i]= d[0][i];
		for(j=1;j<ey-1;j++) a[j][i]= (d[j-1][i]+2*d[j][i]+d[j+1][i])/4;
		a[j][i]= d[j][i];
	}
	free(c); free(d);
#else
	//for(j=0;j<ey;j++)for(i=0;i<ex;i++) a[j][i]= PEX(b);
for(j=0;j<ey-1;j++) { for(i=0;i<ex-1;i++) a[j][i]= PEX(b); a[j][i]= b[j/2][i/2]; }
for(i=0;i<ex-1;i++) a[j][i]= (b[j/2][i/2]+b[j/2][(i+1)/2])/2; a[j][i]= b[j/2][i/2];
#endif
}
void
flat_psf(Dt* m, int len) { int i; for(i=0;i<len;i++) m[i]= 1./len; }
void
#ifdef IIR3
shrink_psf(int h, int v, int h2, int v2, Dt m[V][H], Dt n[V][H]) {
#else
shrink_psf(int h, int v, int h2, int v2, Dt* m0, Dt* n0) {
Dt (*m)[h]= (Dt(*)[h]) m0;
Dt (*n)[h2]= (Dt(*)[h2]) n0;
#endif
int i,j; for(j=0;j<v;j++) for(i=0;i<h;i++) m[j][i]= SPEL(n); norm_hir(h,v,m);
}
#ifdef REAL
void
centroid2(int h, int v, Dt m[][h], Dt t[][h])
// centroid adjuct by linear interpolation
{
	int i,j,k,l,x,y; Dt a,b,d,e, cx= 0, cy= 0, s= 0;

	x= h/2;
	y= v/2;
	a= .001/(h*v);

	for(j=0;j<v;j++)
	for(i=0;i<h;i++) { s+= d= t[j][i]; cx+= d*(i-x); cy+= d*(j-y); m[j][i]= a; }

	cx/= s;
	cy/= s;
	x= (int)cx - (cx<0);
	y= (int)cy - (cy<0);
	d= cx - x;
	e= cy - y;

	for(j=0;j<v;j++) { l= j+y; if(l<0)l= 0; if(l>v-2)l= v-2;
	for(i=0;i<h;i++) { k= i+x; if(k<0)k= 0; if(k>h-2)k= h-2;
		a= t[l  ][k] + d*(t[l  ][k+1] - t[l  ][k]);
		b= t[l+1][k] + d*(t[l+1][k+1] - t[l+1][k]);
		m[j][i]= a + e*(b-a);
	}}
}
#endif

void
expand2(int h, int v, int h2, int v2, Dt m[][h], Dt n[][h2])
{
	int i,j;
#ifdef EXPAND_FIL
	int k,l;
#define Y 1E-3
	Dt s,t,c[v2][h];
	for(j=0;j<v2;j++) {
		i= h&1; if(i) c[j][0]= n[j][0]/2;
		for(k=0;i<h-1;i+=2,k++){
			if(k && k+2<h2)
				t= (5*(n[j][k]+n[j][k+1])-(n[j][k-1]+n[j][k+2]))/8;
			else if(k+1<h2) t= (n[j][k]+n[j][k+1])/2;
			else t= n[j][k]/2;
			s= n[j][k];
			c[j][i  ]= (s<Y)?Y:s;
			c[j][i+1]= (t<Y)?Y:t;
		}
	}
	for(i=0;i<h;i++) {
		j= v&1; if(j) m[0][i]= c[0][i]/2; 
		for(l=0;j<v-1;j+=2,l++){
			if(l && l+2<v2)
				t= (5*(c[l][i]+c[l+1][i])-(c[l-1][i]+c[l+2][i]))/8;
			else if(l+1<v2) t= (c[l][i]+c[l+1][i])/2;
			else t= c[l][i]/2;
			s= c[l][i];
			m[j  ][i]= (s<Y)?Y:s;
			m[j+1][i]= (t<Y)?Y:t;
		}
	}
#else
	//for(j=0;j<v;j++)for(i=0;i<h;i++) m[j][i]= PEX(n);
	for(j=0;j<v-1;j++) { for(i=0;i<h-1;i++) m[j][i]= PEX(n); m[j][i]= n[j/2][i/2]/2; }
	for(i=0;i<h-1;i++) m[j][i]= (n[j/2][i/2]+n[j/2][(i+1)/2])/4; m[j][i]= n[j/2][i/2]/2;
	// psf shoule have soft edge, but image should not.
#endif
}

void
expand_psf(int h, int v, int h2, int v2, Dt* m0, Dt* n0) {
	Dt (*m)[h]= (Dt(*)[h]) m0;
	Dt (*n)[h2]= (Dt(*)[h2]) n0;
#ifdef CENTROID
	int i,j;
	Dt t[v][h];
#ifndef REAL
	int x,y,k,l; Dt d, cx= 0, cy= 0, s= 0;
#endif
#endif
	expand2(h,v,h2,v2,m,n);
#ifdef CENTROID
#ifdef REAL
	for(j=0;j<v;j++) for(i=0;i<h;i++) t[j][i]= m[j][i];
	centroid2(h, v, m, t);
#else
	x= h/2;
	y= v/2;
	for(j=0;j<v;j++)
	for(i=0;i<h;i++) { s+= d= t[j][i]= m[j][i]; cx+= d*(i-x); cy+= d*(j-y); }
	x= cx/s; y= cy/s;
	if(prm) printf("ex_psf: psf shifted (%d, %d) to (0,0).\n",x,y);

	for(j=0;j<v;j++)
	for(i=0;i<h;i++) {
		k= i+x; l= j+y;
		if(k>=0 && k<h && l>=0 && l<v)	m[j][i]= t[l][k];
		else {
			int c= 0;
			Dt a;
			if(k< 0) { c+= -k; k= 0; }
			else if(k>=h) { c+= k-h+1; k= h-1; }
			if(l< 0) { c+= -l; l= 0; }
			else if(l>=v) { c+= l-v+1; l= v-1; }
			a= t[l][k];
			while(c--) a/= 2;
			m[j][i]= a;
		}
	}
#endif
#endif
	norm_hir(h,v,m);
}
//---------------- between Sif image and Dt image array -----------------------
void
sif2dbl(rect *r, int ex, Uc *p0, Dt* b0)
{
	int i,j;
	Uc (*p)[ex]= (Uc(*)[ex]) p0;
	Dt (*b)[ex]= (Dt(*)[ex]) b0;
	for(j=r->s.y;j<r->e.y;j++)
	for(i=r->s.x;i<r->e.x;i++) b[j-r->s.y][i-r->s.x]= gamma_((Dt)p[j][i]);
}
#ifdef GE
void
sif2g_est(int ex, int ey, Uc *p0, Dt b[SY][SX])  // whole sif to double type image
{
Uc (*p)[SX]= (Uc(*)[SX])p0;
int i,j; for(j=0;j<ey;j++) for(i=0;i<ex;i++) b[j][i]= gamma_((Dt)p[j][i]);
}
void
dbl2sif(rect *ir, rect *or, int sx, Dt* b0, Uc* p0)
{
	int i,j,dx,dy;
	Dt (*b)[sx]= (Dt(*)[sx]) b0;
	Uc (*p)[sx]= (Uc(*)[sx]) p0;
	dx= or->s.x - ir->s.x;
	dy= or->s.y - ir->s.y;
	for(j=or->s.y;j<or->e.y;j++)
	for(i=or->s.x;i<or->e.x;i++) p[j][i]= igamma(b[j][i]);
}
#else
void
dbl2sif(rect *ir, rect *or, int px, Dt b[][px], Uc p[][px])
{
	int i,j,dx,dy;
	dx= or->s.x - ir->s.x;
	dy= or->s.y - ir->s.y;
	for(j=0;j<or->e.y - or->s.y;j++)
	for(i=0;i<or->e.x - or->s.x;i++) p[j+ or->s.y][i+ or->s.x]= igamma(b[j+dy][i+dx]);
}
#endif

Dt m1[V][H];
Dt last_err;
Dt lastm[V][H];
Dt last[4][V*H];

#ifdef IIR2
//----------------------- mk_unif --------------------------------
void
mk_unif(rect *or, rect *of, Dt psf[V][H]) // filter estimate by psf
{
	int i,j,k,l,c,cnt= 0; Dt a,e,s;
	int sx= or->s.x; int dx= or->e.x - sx;	// used as pos
	int sy= or->s.y; int dy= or->e.y - sy;	// by es(),in()
	Dt b[V][H];
lo:
	if(psfm==2) clr(m1);
#ifdef BEST
	cp(est0,est);	// est before modified
#endif
	s= c= 0;
	for(j=of->s.y;j<of->e.y;j++)
	for(i=of->s.x;i<of->e.x;i++) {	// flt, err, rat area
		a= 0;
		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++)
			a+= b[l+U][k+M]= best(j-l,i-k)*psf[l+U][k+M];
		if(a< T) a= T; if(a> 1.) a= 1.;
		flt[j][i]= a;
		err[j][i]= in(j,i)-a;
		rat[j][i]= in(j,i)/a;
		s+= err[j][i] * err[j][i]; c++;
		a= err[j][i]/a;
		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++) { e= a*b[l+U][k+M]; es(j-l,i-k)+= e; m1[l+U][k+M]+= e; }
	}
	s/= c;
	if(last_err!=0 && s>last_err && cnt++<4) { avg(psf,lastm); goto lo; }
	else { last_err= s; cpy(lastm,psf); }
	if(pre)printf(" %f\n",sqrt(s));
}

//----------------------- make error ------------------------------------------
void
mk_err(rect *or, rect *of, Dt psf[V][H]) // filter estimate by psf
{
	int i,j,k,l,c=0,cnt= 0;
	int sx= or->s.x; int dx= or->e.x - sx;	// used as pos
	int sy= or->s.y; int dy= or->e.y - sy;	// by es(),in()
	Dt s= 0;
lo:
	s= c= 0;
	for(j=of->s.y;j<of->e.y;j++)
	for(i=of->s.x;i<of->e.x;i++) {	// flt, err, rat area
		Dt a= 0;
		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++) a+= es(j-l,i-k) * psf[l+U][k+M];
		if(a< T) a= T; if(a> 1.) a= 1.;
		flt[j][i]= a;
		err[j][i]= in(j,i)-a;
		rat[j][i]= in(j,i)/a;
		s+= err[j][i] * err[j][i];
		c++;
	}
	s/= c;
	if(last_err!=0 && s>last_err && cnt++<4) { avg(psf,lastm); goto lo; }
	else { last_err= s; cpy(lastm,psf); }
	if(pre)printf(" %f\n",sqrt(s));
}

//---------------------------make estimate image -------------------------------
void
mk_est(rect *or, rect *ir, Dt psf[V][H]) // modify estimate
{
	int i,j,k,l,x1,x2,y1,y2,v; Dt a,e;
	int sx= or->s.x; int dx= or->e.x - sx;	// used as pos of est
	int sy= or->s.y; int dy= or->e.y - sy;
	Dt b[V][H];

	if(psfm==2) clr(m1);
#ifdef BEST
	cp(est0,est);// to separate est before modified
#endif
	if(estm==0) { // alg0
	for(j=ir->s.y;j<ir->e.y;j++)
	for(i=ir->s.x;i<ir->e.x;i++) {
		//est[j][i]+= err[j][i]*0.05;
		a= err[j][i];
		for(l= iy;l<=ay;l++)	// deliver err to est
		for(k= ix;k<=ax;k++) es(j-l,i-k) += a * psf[l+U][k+M];
	}
	/*
	for(j=ir->s.y;j<ir->e.y;j++)
	for(i=ir->s.x;i<ir->e.x;i++) {
		a= 0;
		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++) a+= err[j+l][i+k] * psf[l+U][k+M];
		es(j,i) += a;
	}
	*/
	} else	if(estm==1) { // alg1
	for(j=ir->s.y;j<ir->e.y;j++)
	for(i=ir->s.x;i<ir->e.x;i++) {
		a= 0;
		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++) a+= b[l+U][k+M]= best(j-l,i-k)*psf[l+U][k+M];

		//if(a< T) a= T; if(a> 1.) a= 1.;
		a= err[j][i]/a;
		//a= err[j][i]/flt[j][i];

		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++) {
			e= a*b[l+U][k+M];
			es(j-l,i-k)+= e;
			if(psfm==2) m1[l+U][k+M]+= e;
		}
	}
	} else if(estm==2) { // lucy
	for(j=0;j<dy;j++)
	for(i=0;i<dx;i++) {
		a= 0; if(psfm==2) e= es(j,i);
		for(l= iy;l<=ay;l++)
		for(k= ix;k<=ax;k++) {
			if(j+l>=U && j+l<dy-U && i+k>=M && i+k<dx-M) {
			a+= rat[j+l][i+k] * psf[l+U][k+M];
			if(psfm==2) m1[l+U][k+M]+= e*(rat[j+l][i+k]-1)* psf[l+U][k+M];
			} else a+= psf[l+U][k+M];
		}
		es(j,i)*= a;
	}
	}
	for(j=0;j<dy;j++)
	for(i=0;i<dx;i++) if(es(j,i)<0) es(j,i)= 0;
}

//----------------------- centroid_of_psf --------------------------------------
int
centroid(int sx, int sy, int dx, int dy, Dt m[V][H])
{
	int i,j,x,y; Dt cx=0,cy=0; Dt t[V][H];

	if(norm1(m,(double)1.)) return 1;

	for(j=0;j<V;j++)
	for(i=0;i<H;i++) { Dt d= m[j][i]; cx+= d*(i-M); cy+= d*(j-U); }

	x= cx; y= cy;

	if(x|y){
		printf("cent=(%4.2f,%4.2f)\n",cx,cy);
		cpy(t,m); cp(tmp,est);
		if(y>=0)
		for(j=0;j<V-y;j++) {
			if(x>=0) for(i=0;i<H-x;i++) m[j][i]= t[j+y][i+x];
			else	 for(i=-x;i<H;i++)  m[j][i]= t[j+y][i+x];
		}
		else for(j=-y;j<V;j++) {
			if(x>=0) for(i=0;i<H-x;i++) m[j][i]= t[j+y][i+x];
			else	 for(i=-x;i<H;i++)  m[j][i]= t[j+y][i+x];
		}
		if(y>=0)
		for(j=0;j<dy-y;j++) {
			if(x>=0) for(i=0;i<dx-x;i++) es(j+y,i+x)= tmp[j][i];
			else	 for(i=-x;i<dx;i++)  es(j+y,i+x)= tmp[j][i];
		}
		else for(j=-y;j<dy;j++) {
			if(x>=0) for(i=0;i<dx-x;i++) es(j+y,i+x)= tmp[j][i];
			else	 for(i=-x;i<dx;i++)  es(j+y,i+x)= tmp[j][i];
		}
		//if( norm1(m,(double)1.)) return 1; 
		printf("moved %d %d\n",x,y);
	}
	return 0;
}

void
show_psf(Dt m[V][H])
{
	int k,l;
	for(l=iy;l<=ay;l++)
	for(k=ix;k<=ax;k++) {
Dt d= m[l+U][k+M]*256*50; if(d<0)d=0; if(d>255)d= 255; est[l+U][k+M]= d/255.;
	}
}

//----------------- move psf ---------(estimate psf)----------------------------
int
move_psf(rect *or, rect *of, Dt m[V][H])
{
	int i,j,k,l,c;
	int sx= or->s.x; int dx= or->e.x - sx;
	int sy= or->s.y; int dy= or->e.y - sy;
	Dt a,e=0;
	if(psfm==2){
		for(j=0;j<SY;j++)
		for(i=0;i<SX;i++) e+= best(j,i);
		mls(m,e/accl);
		add(m,m1);
	} else if(psfm) {

	for(l=iy;l<=ay;l++)
	for(k=ix;k<=ax;k++) {
		a= 0;
		for(j=of->s.y;j<of->e.y;j++)
		for(i=of->s.x;i<of->e.x;i++){
#ifdef RECALC
			Dt b= es(j-l,i-k);
#else
			Dt b= best(j-l,i-k);
#endif
			if(psfm==1) e= rat[j][i];
			else e= err[j][i];
			if(e>0){ for(c= 0;c<accl;c++) b*= e; a+= b; }
		}
		m1[l+U][k+M]= a;
	}
	if(psfm==3) sqr(m1);
	if(psfm==4) { sqr(m1); sqr(m1); }
	if(sum(m1)) mlt(m,m1);
	else printf("m1 is all 0\n");
	no_minus(m);

	}
	//centroid(sx,sy,dx,dy,m);
	if(norm1(m,(double)1.)) return 1;
	//show_psf(m);
	if(prm)pr_mat(m);
	return 0;
}

//--------- direction preserve filter ------------------------------
void
dfilter(int th)
{
	int i,j,k,l;
	Dt ul,uc,ur,cl,cc,cr,dl,dc,dr,hd,vd,rd,ld,min,res;
	for(j=1;j<SY-1;j++)
	for(i=1;i<SX-1;i++) {
		ul= est[j-1][i-1];
		uc= est[j-1][i];
		ur= est[j-1][i+1];
		cl= est[j  ][i-1];
		cc= est[j  ][i];
		cr= est[j  ][i+1];
		dl= est[j+1][i-1];
		dc= est[j+1][i];
		dr= est[j+1][i+1];
		hd= fabs(cl+cr-2*cc);
		vd= fabs(uc+dc-2*cc);
		rd= fabs(dl+ur-2*cc);
		ld= fabs(dr+ul-2*cc);
		min= th; res= cc;
		if(hd<min) {min= hd; res= (cl+cr+2*cc)/4; }
		if(vd<min) {min= vd; res= (uc+dc+2*cc)/4; }
		if(rd<min) {min= rd; res= (dl+ur+2*cc)/4; }
		if(ld<min) {min= ld; res= (dr+ul+2*cc)/4; }
		flt[j][i]= res;
	}
	cp(est,flt);
}

//-----------------epsilon filter------------------------------
void
efilter(int th)
{
	int i,j,k,l,cnt;
	Dt c,d,sum;
	for(j=1;j<SY-1;j++)
	for(i=1;i<SX-1;i++) {
		c= est[j][i];
		sum= c;
		cnt= 1;
		for(l=-1;l<=1;l++)
		for(k=-1;k<=1;k++){
			d= est[j+l][i+k];
			if(fabs(d-c)<= th){ sum+= d; cnt++; }
		}
		flt[j][i]= sum/cnt;
	}
	cp(est,flt);
}

void
init_pic(int val, Dt p[SY][SX]){
	int i,j;
	for(j=0;j<SY;j++)
	for(i=0;i<SX;i++) p[j][i]= val;
}

void
write_out_psf_jpg(int lp, char *n, Dt m[V][H],Uc *ot)
{
	write_out(ot,SX*SY,"o.yuv");
	//sprintf(n,"%d.psf",lp); write_psf(n, m);
	sprintf(n,"convert -size %dx%d o.yuv out%d.jpg",SX,SY,lp); system(n);
}

/*
**	iterational inverse filter (IIR2)
*/
int
iir2(int o,int i,int j,rect *or,rect *ir,rect *of,Dt m[V][H],Uc *ot)
{	
	int lp;
	if(norm1(m,(double)1.)) return 1;
#ifndef GE
	sif2dbl(or,SX,&in.y,inp);
	cp(est,inp);
	//init_pic(gamma_((dbt)40.),est);
#endif
	last_err= 1.0;
	for(lp= 0;lp<iloop;lp++){
		if(prm) printf("lp= %d\t",lp);
		if(estm==3)mk_unif(or,of,m);
		else { mk_err(or,of,m); mk_est(or,of,m); }
#ifdef RECALC
		mk_err(or,of,m);
#endif
		if(move_psf(or,of,m)) break;
		//efilter(1);
		dbl2sif(or,ir,SX,(Dt*)est,ot);
		sprintf(name,"o=%d (y=%d,x=%d,it=%4d)", o, j, i, lp);
		wt(ot,name); // display on X-window
		//if(lp%100==0)write_out_psf_jpg(lp,name,m,ot);
	}
	//efilter(1); dbl2sif(or,ir,EX,(Dt*)est,ot);
	//write_out_psf_jpg(lp,name,m,ot);
	return 0;
}
#endif
void
norm(Dt m[V][H], double c) // normalize sum of mat to be 1.
{
	int i,j; double sum= 0;
	for(j=0;j<V;j++) for(i=0;i<H;i++) sum+= m[j][i];
	sum/= c;
	for(j=0;j<V;j++) for(i=0;i<H;i++) m[j][i]/= sum;
}
#define Center (M*H+M)
#define skip_center(i) (i+(i>=Center))
/*
**	blind deconvolution core
*/
#define IE(yy,xx)  est[LM(yy,sy)][LM(xx,sx)]
#define IE0(yy,xx)  IE(yy,xx)
#define IEA(a,b,c) (IE(a,b)=MAX(IE(a,b)+c,T))


#define NH 64
Dt sob[NH];
int cnt[NH];
Dt yl[NH];

Dt nl(Dt a, Dt b,int init)
{
	int i,j;
	static int start= 1;
	Dt t;

	if(init){
		if(start) {
			start= 0;
			for(i=0;i<NH;i++) yl[i]= (Dt)(i+.5)/NH;
		}
		else if(init==2) {
			for(i=0;i<NH;i++) if(cnt[i]) yl[i]= sob[i]/cnt[i];
			t= 1./(1.5*yl[NH-1] - .5*yl[NH-2]);
			for(i=0;i<NH;i++) {
				yl[i]*= t;
				//printf("%d:%f\n",i,yl[i]);
			}
			sob[i]= cnt[i]= 0;
		}
		return 0;
	}
	if(!use_gamma) return a;
	i= a*NH;
	j= a*NH+.5;
	if(i<0) i= 0; if(i>=NH)i= NH-1;
	if(j<0) j= 0; if(j>=NH)j= NH-1;
	sob[j]+= b;
	cnt[j]++;
	a= a*NH-i;
	if(i==NH-1) return yl[i];
	return yl[i]+a*(yl[i+1]-yl[i]);
}

void
iir12(rect *or,int sx,int sy,Dt *i0,Dt *e0,Dt *a0,Dt *s0,int h,int v,Dt **m0,int lay)
{
	int i,j,k,l=lay,c,cnt= 0; Dt a,e,s,t,z;//,u;
	int ix= ixa[l],ax= axa[l],iy= iya[l],ay= aya[l];
	int x= or->s.x; int ex= or->e.x;
	int y= or->s.y; int ey= or->e.y;
	int ln= v*h, flag= 0;
	Dt b[v][h];
	Dt me[v][h]; Dt mf[v][h];
	Dt ce[v][h]; Dt cf[v][h];
	Dt c0[4][v][h];
	Dt (*inp)[sx]= (Dt(*)[sx])i0;
	Dt (*err)[sx]= (Dt(*)[sx])e0;
	Dt (*rat)[sx]= (Dt(*)[sx])a0;
	Dt (*est)[sx]= (Dt(*)[sx])s0;
	Dt (*m)[v][h]= (Dt(*)[v][h])*m0;
#ifndef FULLP
	x+= ax; ex+= ix; y+= ay; ey+= iy;
#endif
lo:	z=s=t=c= 0;
	clrl((Dt*)c0,ln*4);
	nl((Dt)0,(Dt)0,1);

	for(j=y;j<ey;j++) {
		Dt dn= (Dt)j/ey;
		Dt up= 1 - dn;
		clrl((Dt*)ce,ln);
		clrl((Dt*)cf,ln);

		for(l=iy;l<=ay;l++)
		for(k=ix;k<=ax;k++) {
		me[l+v/2][k+h/2]= up*m[0][l+v/2][k+h/2]+dn*m[2][l+v/2][k+h/2];
		mf[l+v/2][k+h/2]= up*m[1][l+v/2][k+h/2]+dn*m[3][l+v/2][k+h/2];
		}

	for(i=x;i<ex;i++) {
		Dt rt= (Dt)i/ex;
		Dt lt= 1 - rt;
		a= 0;
		//s+= est[j][i];

		for(l=iy;l<=ay;l++)
		for(k=ix;k<=ax;k++) a+= b[l+v/2][k+h/2]= IE0(j-l,i-k) *
			(lt*me[l+v/2][k+h/2]+rt*mf[l+v/2][k+h/2]);
		if(a<T) a= T;
		if(a>1) a= 1;
		//if(use_gamma){ if(a>10) a= 1; else a= (exp(2*a)-1)/(exp(2*a)+1); }
#ifdef AG
		z+= pow(inp[j][i],gmm/GMM);
		//z+= inp[j][i];
		s+= a;
		//e= err[j][i]= pow(inp[j][i],gmm/GMM) - a;
		e= err[j][i]= nl(inp[j][i],a,0) - a;
#else
		e= err[j][i]= inp[j][i] - a;
#endif
		rat[j][i]= a= e/a; t+= e*e; c++;
#ifdef Damp
		//if(use_gamma) rat[j][i]= a= (a+1)*(a+1) - 1; //for unif, unif2
		if(use_gamma) rat[j][i]= a*= Damp*fabs(a); //for unif, unif2
		//{u= a+1.; if(u<T) u= T; u= 1000*(inp[j][i]*log(u)-e);
		//if(u<1.) a*= u*u*u*u*u*u*u*u*u*(10-u*9);}
		//if(a<-10)a= -10; //assert(a>-1);
#endif
#ifdef Unified
#define Unif2
#ifdef Unif2
		if(e>0) { flag= 1;
			for(l=iy;l<=ay;l++)
			for(k=ix;k<=ax;k++) {
				Dt r= IE0(j-l,i-k)*e*e; // not good a*a, a or a+TH
				ce[l+v/2][k+h/2]+= r*lt; cf[l+v/2][k+h/2]+= r*rt;
			}
		}
#endif
		for(l=iy;l<=ay;l++)
		for(k=ix;k<=ax;k++) {
			//e= a*b[l+v/2][k+h/2]; IEA(j-l,i-k,e); //c0[l+v/2][k+h/2]+= e;
			if(j-l>=0&&i-k>=0&&j-l<sy&&i-k<sx) {
				est[j-l][i-k]+= a*b[l+v/2][k+h/2];
				if(est[j-l][i-k]<T) est[j-l][i-k]= T;
			}
		}
	}
#ifdef Unif2
	for(l=iy;l<=ay;l++)
	for(k=ix;k<=ax;k++) {
		Dt e0= ce[l+v/2][k+h/2];
		Dt e1= cf[l+v/2][k+h/2];
		c0[0][l+v/2][k+h/2]+= e0*up; c0[1][l+v/2][k+h/2]+= e1*up;
		c0[2][l+v/2][k+h/2]+= e0*dn; c0[3][l+v/2][k+h/2]+= e1*dn;
	}
#endif
	}
#else
	}}
	for(j=y;j<ey;j++) {
		Dt dn= (Dt)j/ey;
		Dt up= 1 - dn;
		for(l=iy;l<=ay;l++)
		for(k=ix;k<=ax;k++) {
		me[l+v/2][k+h/2]= up*m[0][l+v/2][k+h/2]+dn*m[2][l+v/2][k+h/2];
		mf[l+v/2][k+h/2]= up*m[1][l+v/2][k+h/2]+dn*m[3][l+v/2][k+h/2];
		}
	for(i=x;i<ex;i++) {
		Dt rt= (Dt)i/ex;
		Dt lt= 1 - rt;
		a= 1.;
		for(l=iy;l<=ay;l++)
		for(k=ix;k<=ax;k++) a+= 
			(lt*me[l+v/2][k+h/2]+rt*mf[l+v/2][k+h/2])
			* rat[LM(j+l,sy)][LM(i+k,sx)];
		est[j][i]*= a;
	}}
#endif
	t/= c;
	if(last_err!=0 && t>last_err && cnt++<4) {
		avgl((Dt*)m,(Dt*)last,ln*4); if(pre)printf(">"); goto lo;
	}
	last_err= t;
	cpyl((Dt*)last,(Dt*)m,ln*4);
	if(pre)printf("RMSE= %f\n",sqrt(t));
	if(psfm==2) { mlsl((Dt*)c0,accl/s,ln*4); addl((Dt*)m,(Dt*)c0,ln*4); }
	else if(psfm) 
#ifndef Unif2
	for(j=y;j<ey;j++) {
		Dt dn= (Dt)j/ey; Dt up= 1 - dn;
		clrl((Dt*)ce,ln); clrl((Dt*)cf,ln);
		for(i=x;i<ex;i++) {
		Dt rt= (Dt)i/ex; Dt lt= 1 - rt;
		Dt e= err[j][i];
		if(e>0) { flag= 1;
			for(l=iy;l<=ay;l++)
			for(k=ix;k<=ax;k++) {
				Dt r= IE0(j-l,i-k)*e*e;
				ce[l+v/2][k+h/2]+= r*lt; cf[l+v/2][k+h/2]+= r*rt;
			}
		}
		}
		for(l=iy;l<=ay;l++)
		for(k=ix;k<=ax;k++) {
		Dt e0= ce[l+v/2][k+h/2]; Dt e1= cf[l+v/2][k+h/2];
		c0[0][l+v/2][k+h/2]+= e0*up; c0[1][l+v/2][k+h/2]+= e1*up;
		c0[2][l+v/2][k+h/2]+= e0*dn; c0[3][l+v/2][k+h/2]+= e1*dn;
		}
	}
#endif
	if(flag) mltl((Dt*)m,(Dt*)c0,ln*4);

	for(i=0;i<4;i++) norm_hir(h,v,m[i]);
	wtp4(h,v,m);
#ifdef AG
	if(use_gamma){
		if(prm||pre) printf("%f\n",gmm);
		gmm= gmm*sqrt(z/s);
		if(gmm>  5) gmm= 5;
		if(gmm<2.2) gmm= 2.2;
		nl((Dt)0,(Dt)0,2);
	}
#endif
}

void
expands(int l, int dv, int dvl,
	int h,  int v,
	int hu, int vu,
	Dt ml[DY][DX][V][H],
	Dt mu[DY][DX][V][H])
{
	int n,m;

	for(n=0;n<dv;n++)
	for(m=0;m<dv;m++) {
		expand_psf(h,v,hu,vu,(Dt*) ml[n*2][m*2],(Dt*) mu[n][m]);
		cpy(ml[n*2  ][m*2+1],ml[n*2][m*2]);
		cpy(ml[n*2+1][m*2  ],ml[n*2][m*2]);
		cpy(ml[n*2+1][m*2+1],ml[n*2][m*2]);
	}
}

int
read_in_yuv_file(Uc *in, int z, char *n)
{
	int len;
	f= fopen(n, "rb");
	if(!f) { char na[256]; strcpy(na,n); strcat(na,".yuv"); f= fopen(na,"rb"); }
	if(!f) { fprintf(stderr,"%s ropen err\n", n); return 1; }
	if((len= fread(in,1,z,f))<z) { fprintf(stderr,"%s read err len=%d\n",n,len); return 1; }
	fclose(f);
	return 0;
}

void
write_out(Uc *ot, int z,char *n)
{
	f= fopen(n, "wb"); if(!f) { fprintf(stderr,"%s wopen err\n",n); exit(1); }
	if(fwrite(ot,1,z,f)<z) { fprintf(stderr,"%s write err\n",n); exit(1); }
	fclose(f);
}

void
point_rect(Uc *ot, rect* rc)
{
	peak_pos_x= peak_pos_y= 0;
	while(peak_pos_x==0 && peak_pos_y==0) { wt(ot,"point upper left corner"); usleep(100000); }
	rc->s.x= peak_pos_x;
	rc->s.y= peak_pos_y;

	peak_pos_x= peak_pos_y= 0;
	while(peak_pos_x==0 && peak_pos_y==0) { wt(ot,"point lower right corner"); usleep(100000); }
	rc->e.x= peak_pos_x;
	rc->e.y= peak_pos_y;
}

void
sub_rect(rect *a, rect *b) {
	a->s.x-= b->s.x;
	a->s.y-= b->s.y;
	a->e.x-= b->s.x;
	a->e.y-= b->s.y;
}

Dt psfa[DY][DX][V][H];
void
init_psfa(void) {
	int i,j,k,l;
	for(l=0;l<DY;l++)
	for(k=0;k<DX;k++)
	for(j=0;j< V;j++)
	for(i=0;i< H;i++) psfa[l][k][j][i]= (1./(V*H));
}

void
avg_est(int sx, int sy, Dt *est)
{
	int i; Dt a= 0;

	for(i=0;i<sx*sy;i++) a+= est[i];
	a/= sx*sy;
#ifdef FULLP
	for(i=0;i<sx*sy;i++) est[i]= a;
#else
	for(i=ax;i<(sy+iy)*(sx+ix);i++) est[i]= a;
#endif
}

void
average_est(Dt *p0, int l)
{
	Dt *p= p0;
	Dt s= 0;
	int i;
	for(i=0;i<l;i++) s+= *p++;
	p= p0;
	s/= l;
	for(i=0;i<l;i++) *p++= s;
}

Dt le[DY][DX];
void
expand_disp(Dt **sp, int l, rect *or, rect *ir, int ex,int ey, Uc *ot)
{
	for(l--;l>=0;l--) expand_img(ex>>l,ey>>l,sp[l],sp[l+1]);
	dbl2sif(or,ir,ex,sp[0],ot);
}
Dt m3[HIR][DY][DX][V][H];
Dt m3l[DX][V][H];

int mjpg= 1;
Uc psf1[V*6][H*4];
Uc psf4[V*12][H*8];
void
igamma_psf(Dt *m)
{
	int i; Dt *p= m,s= 0;
	for(i=0;i<V*H;i++) { s+= *p= igamma(*p); p++; }
	p= m;
	for(i=0;i<V*H;i++) { *p/= s; p++; }
}
void
mk_psfjpg(Dt *m)
{
	int i,j,k,l;
	Dt *p= m;
	{ Uc *p=(Uc*)psf1; for(i=0;i<sizeof(psf1);i++) *p++= 128; }
	for(j=0;j<V;j++)
	for(i=0;i<H;i++){
		Dt a= *p++;
		int mm= igamma(a*V*H/8);
		for(l=0;l<4;l++)
		for(k=0;k<4;k++) psf1[4*j+l][4*i+k]= mm;
	}
	write_out((Uc*)psf1,sizeof(psf1),"z.yuv");
	sprintf(name,"convert -size %dx%d z.yuv %s_psf.jpg; rm z.yuv",H*4,V*4,fn);
	system(name);
}
void
mk_psfjpg4(Dt *m)
{
	int i,j,k,l,n,ox,oy;
	{ Uc *p=(Uc*)psf4; for(i=0;i<sizeof(psf4);i++) *p++= 128; }
	for(n=0;n<4;n++) { ox= (n&1)?H*4:0; oy= (n&2)?V*4:0;
	for(j=0;j<V;j++)
	for(i=0;i<H;i++){ int d= igamma(*m++*V*H/8);
		for(l=0;l<4;l++)
		for(k=0;k<4;k++) psf4[4*j+l+oy][4*i+k+ox]= d;
	}}
	write_out((Uc*)psf4,sizeof(psf4),"z.yuv");
	sprintf(name,"convert -size %dx%d z.yuv %s_psf.jpg; rm z.yuv",H*8,V*8,fn);
	system(name);
}

//-- deconvolution function --
void
point_get_psf(Uc *in)
{
	peak_pos_x=0;
	while(peak_pos_x==0){ wt(in,"get_psf"); usleep(100000); } avge= 1;
}

void
deconv(int ex, int ey, int sz, Dt m2[V][H], Uc *in, Uc *ot)
{
	int k,x,y;
	rect ir,or;//,of;// of= ir - or;

	if(gpsf) point_get_psf(in);	// get the point of PSF
	else flat_psf((Dt*)m2,V*H);
	norm(m2,(double)1);

#ifdef LPC
	//auto_correlation(m2);
	pr_mat(m2);
	put_mat_sk(m2); inv_mat(N-1,N*2-1);// skip-center setting to N-1 x (2N-1) matrix 
	mult_mat_sk((Dt*)m2,(Dt*)mat3);	// inversed matrix x m2 ==> mat3 (with skipping center)
	prep_mat(mat3);			// to remove 8 point big values
	norm(mat3,(double)-1);
	pr_mat(mat3);
	mat3[U][M]+= 2; pr_mat(mat3); fir(ex,ey,in,ot,mat3);
	wt(ot,"in.yuv");
	getchar();
#endif
#ifdef INV_MAT
	x= y= -1;
	put_mat(m2); inv_mat(N,N*2);
	do{
		if(x!=xoff||y!=yoff) {
			x= xoff; y= yoff;
			sprintf(name,"x= %d, y= %d",x,y);
			get_mat_fir(ex,ey,x,y);
		}
		wt(ot,name);
	}while(1);
#endif
#ifdef Wiener
	gaussian(m2,3);
	pr_mat(m2); or.s.x= or.s.y= 0; or.e.x= ex; or.e.y= ey; ir= or;

	clr_wk(wpsf); set_wk(wpsf,m2); FT(wpsf);
	sif2dbl(&ir,ex,in,inp);
	clr_wk(wk); set_fft(wk,inp); FT(wk);
	divd(wk,wk,wpsf);
	IT(wk); get_fft(wk,est);
	dbl2sif(&or,&ir,ex,est,ot); sprintf(name,"Wiener result"); wt(ot,name);
	getchar();
#endif

#ifdef WienerL
	pr_mat(m2); or.s.x= or.s.y= 0; or.e.x= ex; or.e.y= ey; ir= or;

	clr_wk(wpsf); set_wk(wpsf,m2); FT(wpsf);// PSF is made
	sif2dbl(&ir,ex,in,inp);
	clr_wk(winp); set_fft(winp,inp); 	// inp is made
#define WL1
#ifdef WL1
	FT(winp);			// INP
#else
	cp(west,winp); FT(west);	// EST
#endif
	for(k=0;k<iloop;k++){
	  mult(wk,west,wpsf);		// wk= EST*PSF
#ifdef WL1
	  msub(wk,winp);		// wk= INP - wk
	  if(k&1)mda(wpsf,wk,west,1);	// PSF+= wk/EST
	  else   mda(west,wk,wpsf,1);	// EST+= wk/PSF
	  cp(wk,west); IT(wk); clip_est(wk); get_fft(wk,est);
#else
	  IT(wk); divd(wk,winp,wk); FT(wk); // wk= inp/wk
	  cp(wx,wk); divd(wk,wk,wpsf); divd(wx,wx,west);// WK/= PSF,WX/= EST
	  IT(wk); IT(west); mult(west,wk,west); get_fft(west,est); FT(west); //est*= wk
	  IT(wx); IT(wpsf); mult(wpsf,wx,wpsf); /*mnorm(wpsf);*/ FT(wpsf); //psf*= wx
#endif
	  dbl2sif(&or,&ir,ex,est,ot); sprintf(name,"WienerL %d",k+1); wt(ot,name);
	  printf("%d: %f\n",k+1,wpsf[0][0].re);
	}

#endif

#ifdef IIR1
	{
	int i,l,c, v[5]={V,V2,V3,V4,V5}, h[5]={H,H2,H3,H4,H5};
	Dt *mp[HIR][4], *ip[HIR], *sp[HIR], *ep[HIR], *rp[HIR];
	for(l=0;l<HIR;l++) {
		int s= (ex>>l)*(ey>>l)*sizeof(Dt);
		int p= v[l]*h[l];
		mp[l][0]= malloc(p*sizeof(Dt)*4); for(i=1;i<4;i++) mp[l][i]= mp[l][i-1]+p;
		ip[l]= malloc(s); sp[l]= malloc(s); ep[l]= malloc(s); rp[l]= malloc(s);
	}
	or.s.x= or.s.y= 0;
	or.e.x= ex; or.e.y= ey;
	ir= or;
#ifdef AG
	if(prm||pre)printf("\n");
#endif
	gmm= GMM;
	sif2dbl(&ir,ex,in,ip[0]);
	memcpy(sp[0],ip[0],ex*ey*sizeof(Dt));
	for(l=1;l<HIR;l++) {
		x= ex>>l; y= ey>>l;
		shrink_img(x,y,ip[l],ip[l-1]);
		shrink_img(x,y,sp[l],sp[l-1]);
		if(gpsf) 
		for(i=0;i<4;i++) shrink_psf(h[l],v[l],h[l-1],v[l-1],mp[l][i],mp[l-1][i]);
	}
	if(!gpsf)for(i=0;i<4;i++) flat_psf((Dt*)mp[l-1][i],v[l-1]*h[l-1]);
	while(--l>=0) {
		x= ex>>l; y= ey>>l;
		if(avge)average_est(sp[l],x*y);
		or.e.x= x; or.e.y= y;
		last_err= 0;
		for(k=0;k<((l)?hloop:iloop);k++) {
			sprintf(name,"Est layer:%1d iter=%3d",l,k+1);
			if(prm)printf(" %s ",name);
#ifdef GV
			if(l==0 && k==iloop-GV){
				int i;
				Dt *p= ip[l];
				for(i=0;i<x*y;i++) p[i]= sqrt(p[i]);
				memcpy(sp[l],ip[l],x*y*sizeof(Dt));
				if(avge)average_est(sp[l],x*y);
				gmm= gmm*.5;
				//psfm= 0;
			}
#endif
			iir12(&or,x,y,ip[l],ep[l],rp[l],sp[l],h[l],v[l],(Dt**)mp[l],l);
			expand_disp(sp,l,&or,&ir,ex,ey,ot);
			wt(ot,name);
		}
		if(l)for(i=0;i<4;i++)
			expand_psf(h[l-1],v[l-1],h[l],v[l],mp[l-1][i],mp[l][i]);
	}
	gmm= GMM;
	//if(mjpg)mk_psfjpg((Dt*)mp[0]);
	if(mjpg)mk_psfjpg4((Dt*)mp[0][0]);
	//for(i=0;i<4;i++) igamma_psf((Dt*)mp[0][i]);
	// chroma
	if(HIR>1){
	l= 1; x= ex>>l; y= ey>>l;
	ir.e.x= x; ir.e.y= y; or= ir;
	use_gamma= 0;	// chroma no gamma
	for(c=0;c<2;c++){
		sif2dbl(&ir,x,in+ex*ey+(ex*ey/4)*c,ip[1]);
		memcpy(sp[l],ip[l],x*y*sizeof(Dt));
		if(avge)average_est(sp[l],x*y);
		for(i=0;i<4;i++) shrink_psf(h[l],v[l],h[l-1],v[l-1],mp[l][i],mp[l-1][i]);
		last_err= 0;
		for(k=0;k<cloop;k++) {
			sprintf(name,"Estimate Chroma%d %2d/%2d",c,k+1,cloop);
			if(prm)printf(" %s ",name);
			iir12(&or,x,y,ip[l],ep[l],rp[l],sp[l],h[l],v[l],(Dt**)mp[l],l);
			dbl2sif(&or,&ir,x,sp[l],ot+ex*ey+ex*ey/4*c);
			wt(ot,name);
		}
	}
	}
	use_gamma= 1;
	gmm= GMM;
	//psfm= PSFM;
	for(l=0;l<HIR;l++){
		free(mp[l][0]); free(ip[l]);free(sp[l]);free(ep[l]);free(rp[l]);
	}
	}
#endif

#ifdef IIR2
	init_psfa();
	pr_mat(m2);
#ifdef GE
	sif2g_est(ex,ey,in,inp);
	sif2g_est(ex,ey,in,est);
	if(avge)avg_est(ex,ey,(Dt*)est);
#endif
#ifdef FULL
	or.s.x= or.s.y= 0;
	or.e.x= SX;
	or.e.y= SY;		// or: full image
	ir= or;
	ir.s.x+= M; ir.e.x-= M;
	ir.s.y+= U; ir.e.y-= U;		// ir: inside rect
	of= ir; sub_rect(&of,&or);	// of: ir - or
	iir2(0,0,0,&or,&ir,&of,m2,ot);
#endif
#ifdef PART
	do{
		point_rect(ot,&ir);
		or= ir;
		or.s.x-= M; or.e.x+= M;		// or: outside rect
		or.s.y-= U; or.e.y+= U;
		of= ir; sub_rect(&of,&or);	// of: ir - or
		iir2(0,0,0,&or,&ir,&of,m2,ot);
	}while(1);
#endif
#ifdef DIVD
	{
	Dt m[V][H]; int i,j,o;
	cpy(m,m2); cpy(psfa[0][0],m);

	for(o=0;o<1;o++)
	for(j=0;j<DY;j++)
	for(i=0;i<DX;i++) {
		or.s.x= i*SX/DX; if(i==DX-1) or.e.x= ex; else or.e.x= or.s.x+SX/DX;
		or.s.y= j*SY/DY; if(j==DY-1) or.e.y= ey; else or.e.y= or.s.y+SY/DY;
		ir= or;
#ifdef GE
		if(i==0) ir.s.x+= M; if(i==DX-1) ir.e.x-= M;	// ir: inside of or
		if(j==0) ir.s.y+= U; if(j==DY-1) ir.e.y-= U;	// on the edges of full
#else
		// or: outside of ir, when over edge, it is retrieved, and ir is shrinked
		or.s.x-= M; if(or.s.x< 0) { or.s.x= 0; ir.s.x= M; }
		or.s.y-= U; if(or.s.y< 0) { or.s.y= 0; ir.s.y= U; }
		or.e.x+= M; if(or.e.x>ex) { or.e.x= ex; ir.e.x= ex-M; }
		or.e.y+= U; if(or.e.y>ey) { or.e.y= ey; ir.e.y= ey-U; }
#endif
		if(gpsf)point_get_psf(in);
		else {
			//cpy(m,m2);
			if(j&&i==0) cpy(m,m2);
			if(j) avg(m,psfa[j-1][i]);
			if(o){
				if(i<DX-1) avg(m,psfa[j][i+1]);
				if(j<DY-1) avg(m,psfa[j+1][i]);
				avg(m,psfa[ j][ i]);
			}
		}
		of= ir; sub_rect(&of,&or);	// of: ir - or
		if(iir2(o,i,j,&or,&ir,&of,m,ot)) break ;
		//centroid(0,0,0,0,m);
		cpy(psfa[j][i],m);
	}
	}	
#endif
#endif
#ifdef IIR3
	{
	int l, v[5]={V,V2,V3,V4,V5}, h[5]={H,H2,H3,H4,H5};
	Dt *ip[HIR], *sp[HIR], *ep[HIR], *rp[HIR];
	Dt *p; int i,e,x,y;
	int c,m,n;
	int xw= ex>>(HIR-1);
	int yw= ey>>(HIR-1);
	int dv= 1;
	for(l=0;l<HIR;l++) {
		int s= (ex>>l)*(ey>>l)*sizeof(Dt);
		ip[l]= malloc(s); sp[l]= malloc(s); ep[l]= malloc(s); rp[l]= malloc(s);
	}
	or.s.x= or.s.y= 0;
	or.e.x= ex; or.e.y= ey;
	ir= or;
	sif2dbl(&ir,ex,in,ip[0]);
	memcpy(sp[0],ip[0],ex*ey*sizeof(Dt));
	cpyl((Dt*)m3[0][0][0],(Dt*)m2,V*H);

	for(l=1;l<HIR;l++) {
		x= ex>>l; y= ey>>l;
		shrink_img(x,y,ip[l],ip[l-1]);
		shrink_img(x,y,sp[l],sp[l-1]);
		if(gpsf) shrink_psf(h[l],v[l],h[l-1],v[l-1],m3[l][0][0],m3[l-1][0][0]);
	}
	if(!gpsf) flat_psf((Dt*)m3[l-1][0][0],v[l-1]*h[l-1]);

	while(--l>=0) { // hierarchical divisional (up to down)
		int m,n,d;
		int ln= v[l]*h[l];
		x= ex>>l; y= ey>>l;
		if(avge) average_est(sp[l],x*y);
		for(k=0;k<((l)?hloop:iloop) ;k++)
		for(or.s.y= n= 0; n<dv; n++, or.s.y+= yw) 
		for(or.s.x= m= 0; m<dv; m++, or.s.x+= xw) {
			sprintf(name,"Est layer:%2d iter=%3d (%2d, %2d)",l,k+1,n,m);
			if(prm)printf(" %s ",name);
			or.e.x= or.s.x + xw;
			or.e.y= or.s.y + yw;
			if(!k)le[n][m]= 0;
			last_err= le[n][m];
			iir12(&or,x,y,ip[l],ep[l],rp[l],sp[l],h[l],v[l],(Dt*)m3[l][n][m],l);
			le[n][m]= last_err;
			expand_disp(sp,l,&or,&ir,ex,ey,ot);
			wt(ot,name);
		}
		/*
		if(l<HIR-1) // rearrange PSFs
		for(n=0;n<dv;n++)
		for(m=0;m<dv;m++) {
			int c=0; Dt t[V][H];
			clr(t);
			if(m >  0){ addl((Dt*)t,(Dt*)m3[l][n][m-1],ln); c++; }	// left
			if(n >  0){ addl((Dt*)t,(Dt*)m3[l][n-1][m],ln); c++; cpy(m3[l][n-1][m],m3l[m]); } // up
			if(m+1<dv){ addl((Dt*)t,(Dt*)m3[l][n][m+1],ln); c++; }	// right
			if(n+1<dv){ addl((Dt*)t,(Dt*)m3[l][n+1][m],ln); c++; }	// down
			cpy(m3l[m],m3[l][n][m]);		// psf to lb
			intpl((Dt*)m3l[m],(Dt*)t,0.75/c,ln);
			if(prm)printf("rearrange psf %d(%d %d)\n",l,n,m);
			norm_hir(h[l],v[l],m3l[m]);
		}
		*/
		if(l) expands(l, dv, 2, h[l-1], v[l-1], h[l], v[l], m3[l-1], m3[l]);
		dv*= 2;
	}
	// for chroma
	l= 1;
	dv/= 4;
	x= ex>>l; y= ey>>l;
	ir.e.x= x; ir.e.y= y;
	or= ir;
	//psfm_bak= psfm;
	use_gamma= 0;	// chroma no gamma
	for(c=0;c<2;c++){
		sif2dbl(&ir,x,in+ex*ey+ex*ey/4*c,ip[l]);
		memcpy(sp[l],ip[l],x*y*sizeof(Dt));
		if(avge) average_est(sp[l],x*y); //<- this is average of all image
		for(n= 0; n<dv; n++) 
		for(m= 0; m<dv; m++)
			shrink_psf(h[l],v[l],h[l-1],v[l-1],m3[l][n][m],m3[l-1][n][m]);

		for(k=0;k<cloop;k++)
		for(or.s.y= n= 0; n<dv; n++, or.s.y+= yw) 
		for(or.s.x= m= 0; m<dv; m++, or.s.x+= xw) {
			sprintf(name,"Estimate Chroma%d %2d/%2d",c,k+1,cloop);
			if(prm)printf(" %s ",name);
			last_err= 0;
			or.e.x= or.s.x + xw;
			or.e.y= or.s.y + yw;
			iir12(&or,x,y,ip[l],ep[l],rp[l],sp[l],h[l],v[l],(Dt*)m3[l][n][m],l);
			dbl2sif(&or,&ir,x,sp[l],ot+ex*ey+ex*ey/4*c);
			wt(ot,name);
		}
	
	}
	use_gamma=1;
	for(l=0;l<HIR;l++){ free(ip[l]);free(sp[l]);free(ep[l]);free(rp[l]); }
	}
#endif
}

void
cut_out(char *n,int sx,int sy, int d,int e, Uc *ot)
{
	int j,c,ex= sx+d,ey= sy+e;
	FILE *f= fopen(n, "wb"); if(!f) { fprintf(stderr,"%s wo err\n",n); exit(1); }
	for(j=sy;j<ey;j++)
	 if(fwrite(ot+j*ex+sx,1,d,f)<d) { fprintf(stderr,"%s wr err\n",n); exit(1); }
	sx/=2; sy/=2; ey/=2; d/=2;
	for(c=0;c<2;c++) for(j=sy;j<ey;j++)
	 if(fwrite(ot+ex*ey+j*(ex/2)+sx,1,d,f)<d) { fprintf(stderr,"%s wr err\n",n); exit(1); }
	fclose(f);
}


void
get_psf(int x, int y,Uc *in,int ex)
{
	int i,j;
	for(j=0;j<V;j++)
	for(i=0;i<H;i++) mat2[j][i]= gamma_((Dt)in[(y+j-U)*ex+x+i-M]);
	norm1(mat2,(double)1.);
	printf("got_psf: at (%d,%d).\n",x,y);
}

